//=========================================================================--
// This file is a part of the OpenRISC 1001 VHDL model
// Copyright (C) 2000 Damjan Lampret and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
//---------------------------------------------------------------------------
// IP core: OpenRISC 1001, a 32-bit RISC processor core from OR1K family.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * optimized for size (without MMU and Caches)
//  * optimized for Xilinx Virtex/Spartan2 FPGAs (over 100MHz in -6 parts)
// Documentation: http://www.opencores.org/cores/or1k/or1001/
//---------------------------------------------------------------------------
// File: ram16xNd.vhd
// Entity: ram16xNd
// Author: Damjan Lampret, lampret@opencores.org
// Description:	
//  Arithmetic and Logic Unit.
// Architecture Notes:	
//  Add/subtract and bitwise logical operations in separate processes.
//  Shift and rotate in two linked processes. Compare in a third pair of
//  processes.
// TO DO:
//
//---------------------------------------------------------------------------
// Revision History
// $Log: alu.vhd,v $
// Revision 1.1.1.1  2000/04/18 13:19:31  lampret
// This implementation is now in the middle of modifications to have one CPU core both for OpenRISC 1000 architecture and MIPS I.
//
// Translation to Verilog by Peter Monta <pmonta@terayon.com>, 22 April 2000

`include "config.h"

module alu(a, b, lop, aop, cop, save_flag,
  sum, logic, shifted, c_flag, f_flag, rst, clk);

  parameter datawidth = `DATA_WIDTH;

  input [datawidth-1:0] a;
  input [datawidth-1:0] b;
  input [1:0] lop;
  input aop;
  input [3:0] cop;
  input save_flag;
  output [datawidth-1:0] sum;
  output [datawidth-1:0] logic;
  output [datawidth-1:0] shifted;
  output c_flag;
  output f_flag;
  input rst;
  input clk;

  reg [datawidth-1:0] logic;
  reg [datawidth-1:0] shifted;
  reg f_flag;

// signals related to magnitude compare
  wire a_eq_b;
  wire a_lt_b;

// signals related to shift
  reg [2*datawidth-1:0] shifted1;
  wire [1:0] sh_amount1;
  wire [4:2] sh_amount2;

//
// Add and Subtract operations
//
  assign sum = (aop==`AOP_ADD) ? a+b : a-b;

//
// Bit-wise logic operations
//
  always @(lop or a or b)
    case (lop)
      `LOP_AND: logic = a & b;
      `LOP_OR: logic = a | b;
      `LOP_XOR: logic = a ^ b;
      `LOP_NOR: logic = ~(a | b);
    endcase

  assign sh_amount1 = ~b[1:0];
//
// Shift and rotate operations (first part)
//
  reg [2*datawidth-1:0] shifting;

  always @(posedge clk or posedge rst)
    if (rst)
      shifting = #1 0;
    else begin
      shifting[datawidth-1:0] = a;
      shifting[2*datawidth-1:datawidth] = 0;
      case (lop)
        `SH_LEFTLOGIC:
          begin
            shifting[datawidth-1:0] = 0;
            shifting[2*datawidth-1:datawidth-1] = {1'b0,a};
          end
        `SH_RIGHTARITH:
          begin
            shifting[2*datawidth-1:datawidth] = {datawidth{a[datawidth-1]}};
          end
        `ROT_LEFT:
          begin
            shifting[datawidth-2:0] = a[datawidth-2:0];
            shifting[2*datawidth-1:datawidth-1] = {1'b0,a};
          end
      endcase
      if (sh_amount1[1])
        shifting[datawidth+1:0] = shifting[datawidth+3:2];
      if (sh_amount1[0])
        shifting[datawidth-1:0] = shifting[datawidth:1];
      shifted1 <= #1 shifting;
    end

  assign sh_amount2 = ~b[4:2];
//
// Shift and rotate operations (second part)
//
  reg [2*datawidth-1:0] shifting_;

  always @(shifted1 or sh_amount2) begin
    shifting_ = shifted1;
    if (sh_amount2[4])
      shifting_[47:0] = shifting_[2*datawidth-1:datawidth/2];
    if (sh_amount2[3])
      shifting_[39:0] = shifting_[47:datawidth/4];
    if (sh_amount2[2])
      shifting_[35:0] = shifting_[39:datawidth/8];
    shifted = shifting_[datawidth-1:0];
  end

//
// Magnitude compare operations
//
  wire [datawidth-1:0] a_unsigned = {a[datawidth-1] ^ cop[`SIGNED_COMPARE], a[datawidth-2:0]};
  wire [datawidth-1:0] b_unsigned = {b[datawidth-1] ^ cop[`SIGNED_COMPARE], b[datawidth-2:0]};
  assign a_eq_b = (a_unsigned==b_unsigned);
  assign a_lt_b = (a_unsigned<b_unsigned);

//
// Set compare flags
//
  always @(posedge clk or posedge rst)
    if (rst)
      f_flag <= #1 0;
    else
      if (save_flag)
        case (cop[2:0])
          `COP_SFEQ: f_flag <= #1 a_eq_b;
          `COP_SFNE: f_flag <= #1 !a_eq_b;
          `COP_SFGT: f_flag <= #1 !(a_eq_b || a_lt_b);
          `COP_SFGE: f_flag <= #1 !a_lt_b;
          `COP_SFLT: f_flag <= #1 a_lt_b;
          `COP_SFLE: f_flag <= #1 a_eq_b || a_lt_b;
        endcase

endmodule
