//=========================================================================--
// This file is a part of the OpenRISC 1001 VHDL model
// Copyright (C) 2000 Damjan Lampret and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
//---------------------------------------------------------------------------
// IP core: OpenRISC 1001, a 32-bit RISC processor core from OR1K family.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * optimized for size (without MMU and Caches)
//  * optimized for Xilinx Virtex/Spartan2 FPGAs (over 100MHz in -6 parts)
// Documentation: http://www.opencores.org/cores/or1k/or1001/
//---------------------------------------------------------------------------
// File: biu.vhd
// Entity: biu
// Author: Damjan Lampret, lampret@opencores.org
// Description:	
//  Bus interface unit that supports single and burst transactions.
// Architecture Notes:	
//  
// TO DO:
//  * this new version was not tested with Joon Lee's SDRAM controller (yet)
//  * transactions wider than external port should be split into several
//    narrower transactions by BIU.
//---------------------------------------------------------------------------
// Revision History
// $Log: biu.vhd,v $
// Revision 1.1.1.1  2000/04/18 13:19:31  lampret
// This implementation is now in the middle of modifications to have one CPU core both for OpenRISC 1000 architecture and MIPS I.
//
// Translation to Verilog by Peter Monta <pmonta@terayon.com>, 22 April 2000

`include "config.h"

module biu(ext_addr, ext_load, ext_store, ext_be, ext_rw, ext_burstn,
  ext_validn, ext_busyn, ext_irqn, ext_csn, ext_brn, ext_bgn,
  fetch_addr, fetch_data, fetch_valid, fetch_biubusy,
  lsu_addr, lsu_load, lsu_store, lsu_size, lsu_validload, lsu_validstore, lsu_stall,
  biu_stall,
  rst, clk);

  parameter burst_bits = 4;

// external native bus interface
  output [`PADDR_WIDTH-1:0] ext_addr;
  input [`DATA_WIDTH-1:0] ext_load;
  output [`DATA_WIDTH-1:0] ext_store;
  output [1:0] ext_be;
  output ext_rw;
  output ext_burstn;
  output ext_validn;
  input ext_busyn;
  output [`IRQS-1:0] ext_irqn;
  output [`CSS-1:0] ext_csn;
  output ext_brn;
  input ext_bgn;
// fetch unit signals
  input [`LADDR_WIDTH-1:0] fetch_addr;
  output [`DATA_WIDTH-1:0] fetch_data;
  input fetch_valid;
  output fetch_biubusy;
// load/store unit signals
  input [`LADDR_WIDTH-1:0] lsu_addr;
  output [`DATA_WIDTH-1:0] lsu_load;
  input [`DATA_WIDTH-1:0] lsu_store;
  input [1:0] lsu_size;
  input lsu_validload;
  input lsu_validstore;
  output lsu_stall;
// biu
  input biu_stall;
// reset & clock
  input rst;
  input clk;

  reg [`PADDR_WIDTH-1:0] ext_addr;
  reg [`DATA_WIDTH-1:0] ext_store;
  reg ext_rw;
  reg ext_validn;
  reg [`CSS-1:0] ext_csn;
  reg [`DATA_WIDTH-1:0] lsu_load;

  reg fetching;

  assign ext_brn = ! (lsu_validload || lsu_validstore || fetch_valid);

  always @(posedge clk or posedge rst)
    if (rst)
      lsu_load <= #1 0;
    else
      if (lsu_validload)
        lsu_load <= #1 ext_load;

  always @(posedge clk or posedge rst)
    if (rst)
      fetching <= #1 0;
    else
      fetching <= #1 !lsu_validload && !lsu_validstore;

  assign fetch_data = fetching ? ext_load : {`OR32_NOP,26'b00000000000000000000000000};

  always @(lsu_addr or ext_busyn or lsu_store or fetch_addr or ext_load or lsu_validload or lsu_validstore or fetch_valid)
    if (lsu_validload) begin
      ext_addr = lsu_addr;
      ext_validn = 0;
      ext_csn = 1;
      ext_rw = 1;
    end else if (lsu_validstore && biu_stall) begin
      ext_addr = lsu_addr;
      ext_store = lsu_store;
      ext_validn = 0;
      ext_csn = 1;
      ext_rw = 0;
    end else if (fetch_valid) begin
      ext_addr = fetch_addr;
      ext_validn = 0;
      ext_csn = 1;
      ext_rw = 1;
    end else begin
      ext_addr = 0;
      ext_rw = 1;
      ext_validn = 1;
    end

  assign fetch_biubusy = !ext_busyn;
  assign lsu_stall = !ext_busyn;

endmodule
