//=========================================================================--
// This file is a part of the OpenRISC 1001 VHDL model
// Copyright (C) 2000 Damjan Lampret and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
//---------------------------------------------------------------------------
// IP core: OpenRISC 1001, a 32-bit RISC processor core from OR1K family.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * optimized for size (without MMU and Caches)
//  * optimized for Xilinx Virtex/Spartan2 FPGAs (over 100MHz in -6 parts)
// Documentation: http://www.opencores.org/cores/or1k/or1001/
//---------------------------------------------------------------------------
// File: cpu.vhd
// Entity: cpu
// Author: Damjan Lampret, lampret@opencores.org
// Description:	
//  Instantiation of all CPU subcomponents.
// Architecture Notes:	
//  
// TO DO:
//  
//---------------------------------------------------------------------------
// Revision History
// $Log: cpu.vhd,v $
// Revision 1.1.1.1  2000/04/18 13:19:34  lampret
// This implementation is now in the middle of modifications to have one CPU core both for OpenRISC 1000 architecture and MIPS I.
//
// Translation to Verilog by Peter Monta <pmonta@terayon.com>, 22 April 2000

`include "config.h"

module cpu(addr, datain, dataout, be,
  rw, burstn, validn, busyn,
  irqn, csn,
  brn, bgn,
  rstn, clk);

  output [`PADDR_WIDTH-1:0] addr;
  input [`DATA_WIDTH-1:0] datain;
  output [`DATA_WIDTH-1:0] dataout;
  output [1:0] be;  //FIXME: width
  output rw;
  output burstn;
  output validn;
  input busyn;
  output [`IRQS-1:0] irqn;
  output [`CSS-1:0] csn;
  output brn;
  input bgn;
  input rstn;
  input clk;

  reg [`DATA_WIDTH-1:0] dataout;

// signals and variables
  wire [`LADDR_WIDTH-1:0] lsu_addr;
  wire [`DATA_WIDTH-1:0] lsu_load;
  wire [`DATA_WIDTH-1:0] lsu_store;
  wire [1:0] lsu_size;
  wire lsu_validload;
  wire lsu_validstore;
  wire lsu_stall;

  reg [`DATA_WIDTH-1:0] biu_load;
  wire [`DATA_WIDTH-1:0] biu_store;
  wire biu_rw;

  wire [`LADDR_WIDTH-1:0] spr_data;
  wire fetcher_cs;
  wire spr_write;
  wire spr_addr;

  wire exception;
  wire [3:0] except_type;

  wire iu_stall;

// signals related to instruction fetch
  wire [`LADDR_WIDTH-1:0] fetch_addr;
  wire [`DATA_WIDTH-1:0] fetch_data;
  wire fetch_stall;
  wire fetch_biubusy;
  wire fetch_branch;
  wire fetch_valid;

// signals that control instruction decode stage of integer data path
  wire [3:0] rfaddr1;
  wire [3:0] rfaddr2;

// signals that control execute stage of integer data path
  wire [`DATA_WIDTH-1:0] imm;
  wire [1:0] sel1;
  wire [1:0] sel2;
  wire [1:0] lop;
  wire [3:0] cop;
  wire aop;

// signals that control write-back stage of integer data path
  wire rfenw;
  wire [1:0] selwb;
  wire [3:0] rfaddrw;

// signals that control write-back stage of integer data path
  wire save_flag;
  wire f_flag;

  wire rst = !rstn;

// component instances
  biu #(4) biu_inst(
    addr, biu_load, biu_store, be, biu_rw, burstn, validn, busyn, irqn,
    csn, brn, bgn,
    fetch_addr, fetch_data, fetch_valid, fetch_biubusy,
    lsu_addr, lsu_load, lsu_store, lsu_size, lsu_validload, lsu_validstore, lsu_stall,
    fetch_stall,
    rst, clk);

  fetcher fetcher_inst(
    spr_data, fetcher_cs, spr_write, spr_addr,
    fetch_data, fetch_addr, fetch_valid,
    fetch_branch, lsu_store[`LADDR_WIDTH-1:0],
    fetch_stall, exception, except_type,
    rst, clk);

  iu iu_inst(
    imm, sel1, sel2, lop, cop, aop,
    rfenw, save_flag, f_flag, selwb, rfaddrw, rfaddr1, rfaddr2,
    lsu_load, lsu_store, lsu_addr, iu_stall,
    rst, clk);

  iuctrl iuctrl_inst(
    fetch_data, fetch_stall, fetch_biubusy, fetch_branch,
    rfaddr1, rfaddr2,
    imm, sel1, sel2, lop, cop, aop,
    save_flag, f_flag, rfenw, selwb, rfaddrw,
    exception, except_type,
    lsu_size, lsu_validload, lsu_validstore, lsu_stall, iu_stall,
    rst, clk);

  always @(datain or biu_store or biu_rw) begin
//    if biu_rw = '0' then
    dataout <= biu_store;
//    else
    biu_load <= datain;
  end

  assign rw = biu_rw;

endmodule
