//=========================================================================--
// This file is a part of the OpenRISC 1001 VHDL model
// Copyright (C) 2000 Damjan Lampret and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
//---------------------------------------------------------------------------
// IP core: OpenRISC 1001, a 32-bit RISC processor core from OR1K family.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * optimized for size (without MMU and Caches)
//  * optimized for Xilinx Virtex/Spartan2 FPGAs (over 100MHz in -6 parts)
// Documentation: http://www.opencores.org/cores/or1k/or1001/
//---------------------------------------------------------------------------
// File: fetcher.vhd
// Entity: fetcher
// Author: Damjan Lampret, lampret@opencores.org
// Description:	
//  OR32 fetch unit including exception PC and branch EA precalculation.
// Architecture Notes:	
//  
// TO DO:
//  * 
//---------------------------------------------------------------------------
// Revision History
// $Log: fetcher.vhd,v $
// Revision 1.1.1.1  2000/04/18 13:19:35  lampret
// This implementation is now in the middle of modifications to have one CPU core both for OpenRISC 1000 architecture and MIPS I.
//
// Translation to Verilog by Peter Monta <pmonta@terayon.com>, 22 April 2000

`include "config.h"

module fetcher(spr_data, spr_cs, spr_write, spr_addr,
  fetched, fetch_addr, fetch_valid,
  branch, branch_offset,
  stall, exception, except_type,
  rst, clk);

  output [`LADDR_WIDTH-1:0] spr_data;
  input spr_cs;
  input spr_write;
  input spr_addr;
  input [`INSN_WIDTH-1:0] fetched;
  output [`LADDR_WIDTH-1:0] fetch_addr;
  output fetch_valid;
  input branch;
  input [`LADDR_WIDTH-1:0] branch_offset;
  input stall;
  input exception;
  input [3:0] except_type;
  input rst;
  input clk;

  reg [`LADDR_WIDTH-1:0] spr_data;
  reg [`LADDR_WIDTH-1:0] fetch_addr;
  reg fetch_valid;

// signals and shared variables
  reg [`LADDR_WIDTH-1:0] lr_reg;
  reg [`LADDR_WIDTH-1:0] except_reg;

//
// Calculate next EA for fetching instruction
//

  wire [2:0] decision = {stall, branch, exception};
  reg [`LADDR_WIDTH-1:0] pc_reg;
  reg [`LADDR_WIDTH-1:0] stallpc_reg;

  always @(posedge clk or posedge rst)
    if (rst) begin
      pc_reg <= #1 0;
      stallpc_reg <= #1 0;
      fetch_addr <= #1 0;
      fetch_valid <= #1 0;
    end else begin
$display($time,,"dec %x %x %x",stall,branch,exception);
      fetch_valid <= #1 1;
      case (decision)
        3'b000: begin
                  stallpc_reg <= #1 pc_reg;
                  pc_reg <= #1 pc_reg + 4;
$display($time,,"incr pc");
                end
        3'b001: begin
                  except_reg <= #1 pc_reg;
                  stallpc_reg <= #1 pc_reg;
                  pc_reg <= #1 {4'b0000,except_type,2'b00};
                end
        3'b010: begin
                  lr_reg <= #1 pc_reg;
                  stallpc_reg <= #1 pc_reg;
                  pc_reg <= #1 pc_reg + branch_offset;
                end
        3'b011: begin
                  except_reg <= #1 pc_reg;
                  stallpc_reg <= #1 pc_reg;
                  pc_reg <= #1 {4'b0000,except_type,2'b00};
                end
        default: begin
$display($time,,"decision: %x fetcher: default",decision);
                   pc_reg <= #1 stallpc_reg;
                   except_reg <= #1 except_reg;
                   lr_reg <= #1 lr_reg;
                 end
      endcase
      fetch_addr <= #1 pc_reg;
    end

//
// Read/write except_reg and lr_reg
//
  wire [2:0] sdecision = { spr_cs, spr_write, spr_addr };

  always @(posedge clk or posedge rst)
    if (rst) begin
//      except_reg <= #1 0;
//      lr_reg <= #1 0;
    end else
      case (decision)
        3'b100: spr_data <= #1 lr_reg;
        3'b101: spr_data <= #1 except_reg;
//        3'b110: lr_reg <= #1 spr_data;
//        3'b111: except_reg <= #1 spr_data;
        default: spr_data <= 0;
      endcase

endmodule
