//=========================================================================--
// This file is a part of the OpenRISC 1001 VHDL model
// Copyright (C) 2000 Damjan Lampret and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
//---------------------------------------------------------------------------
// IP core: OpenRISC 1001, a 32-bit RISC processor core from OR1K family.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * optimized for size (without MMU and Caches)
//  * optimized for Xilinx Virtex/Spartan2 FPGAs (over 100MHz in -6 parts)
// Documentation: http://www.opencores.org/cores/or1k/or1001/
//----------------------------------------------------------------------------
// File: iu.vhd
// Entity: iu
// Author: Damjan Lampret, lampret@opencores.org
// Description:	
//  OR1001 Integer data path. Pipeline stages are fetch, decode, issue,
//  execute, writeback and retire.
// Architecture Notes:	
//  Pipeline stages are separated by process 'stages'.
// TO DO:
//  * 
//---------------------------------------------------------------------------
// Revision History
// $Log: iu.vhd,v $
// Revision 1.1.1.1  2000/04/18 13:19:36  lampret
// This implementation is now in the middle of modifications to have one CPU core both for OpenRISC 1000 architecture and MIPS I.
//
// Translation to Verilog by Peter Monta <pmonta@terayon.com>, 22 April 2000

`include "config.h"

module iu(imm, sel1, sel2, lop, cop, aop,
  rfenw, save_flag, f_flag, selwb, rfaddrw, rfaddr1, rfaddr2,
  lsu_load, lsu_store, lsu_ea,
  stall,
  rst, clk);

  parameter datawidth = 32;

  input [datawidth-1:0] imm;
  input [1:0] sel1;
  input [1:0] sel2;
  input [1:0] lop;
  input [3:0] cop;
  input aop;
  input rfenw;
  input save_flag;
  output f_flag;
  input [1:0] selwb;
  input [3:0] rfaddrw;
  input [3:0] rfaddr1;
  input [3:0] rfaddr2;
  input [datawidth-1:0] lsu_load;
  output [datawidth-1:0] lsu_store;
  output [`LADDR_WIDTH-1:0] lsu_ea;
  input stall;
  input rst;
  input clk;

  reg [`LADDR_WIDTH-1:0] lsu_ea;

// decode stage related signals
  wire [datawidth-1:0] id1_rf, id2_rf;

// issue stage related signals
  reg [datawidth-1:0] is1, is2;

// execute stage related signals
  reg [datawidth-1:0] ex1_reg, ex2_reg;
  wire [datawidth-1:0] ex_arith, ex_logic, ex_shift;

// writeback stage related signals
  reg [datawidth-1:0] wb_muxed;
  reg [datawidth-1:0] wb_arith_reg, wb_logic_reg, wb_shift_reg;

// retire stage related signals
  reg [datawidth-1:0] ret_muxed;

//
// Model datapath pipeline registers
//
  always @(posedge clk or posedge rst)
    if (rst) begin
      ret_muxed <= #1 0;
      wb_arith_reg <= #1 0;
      wb_logic_reg <= #1 0;
      wb_shift_reg <= #1 0;
      ex1_reg <= #1 0;
      ex2_reg <= #1 0;
    end else
      if (stall==0) begin
        ret_muxed <= #1 wb_muxed;
        wb_arith_reg <= #1 ex_arith;
        wb_logic_reg <= #1 ex_logic;
        wb_shift_reg <= #1 ex_shift;
        ex1_reg <= #1 is1;
        ex2_reg <= #1 is2;
      end

//
// Register File
//
  regfile #(datawidth) rf(rfaddrw, rfaddr1, rfaddr2,
    wb_muxed, id1_rf, id2_rf, rfenw, rst, clk);

//
// Issue stage
//

// Mux operand sources for operand 1

  always @(sel1 or id1_rf or wb_muxed or ret_muxed)
    case (sel1)
      `SEL1_REGFILE: is1 = id1_rf;
      `SEL1_FORW_WB: is1 = wb_muxed;
      `SEL1_FORW_RET: is1 = ret_muxed;
      default: is1 = 0;
    endcase

// Mux operand sources for operand 2

  always @(sel2 or id2_rf or imm or wb_muxed or ret_muxed)
    case(sel2)
      `SEL2_REGFILE: is2 = id2_rf;
      `SEL2_IMM: is2 = imm;
      `SEL2_FORW_WB: is2 = wb_muxed;
      `SEL2_FORW_RET: is2 = ret_muxed;
      default: ;
    endcase

//
// Write-back
//

// Mux outputs from all funtional units to one 'wb_mux' bus

  always @(selwb or wb_arith_reg or wb_logic_reg or wb_shift_reg or lsu_load)
    case (selwb)
      `SELWB_ARITH: wb_muxed = wb_arith_reg;
      `SELWB_LOGIC: wb_muxed = wb_logic_reg;
      `SELWB_SH_ROT: wb_muxed = wb_shift_reg;
      `SELWB_LSU: wb_muxed = lsu_load;
      default: wb_muxed = 0;
    endcase

  wire open;

  alu #(`DATA_WIDTH) alu_inst(ex1_reg, ex2_reg, lop, aop, cop,
    save_flag, ex_arith, ex_logic, ex_shift, open, f_flag,
    rst, clk);

//lsu_ea <= wb_arith_reg (LADDR_WIDTH-1 downto 0);
  reg [datawidth-1:0] lsu_imm;

  always @(posedge clk or posedge rst)
    if (rst) begin
      lsu_ea <= #1 0;
      lsu_imm <= #1 0;
    end else begin
//      lsu_ea <= #1 ex_arith[`LADDR_WIDTH-1:0];
      lsu_ea <= #1 ex1_reg + lsu_imm;
      lsu_imm <= #1 imm; //fixme: := vs <=
    end

  assign lsu_store = ex2_reg;
    
endmodule
