//=========================================================================--
// This file is a part of the OpenRISC 1001 VHDL model
// Copyright (C) 2000 Damjan Lampret and OPENCORES.ORG
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// See the file COPYING for the full details of the license.
//
// OPENCORES.ORG is looking for new open source IP cores and developers that
// would like to help in our mission.
//---------------------------------------------------------------------------
// IP core: OpenRISC 1001, a 32-bit RISC processor core from OR1K family.
// Features:
//  * fast exception processing (see architecture manual for details)
//  * optimized for size (without MMU and Caches)
//  * optimized for Xilinx Virtex/Spartan2 FPGAs (over 100MHz in -6 parts)
// Documentation: http://www.opencores.org/cores/or1k/or1001/
//----------------------------------------------------------------------------
// File: iuctrl.vhd
// Entity: iuctrl
// Author: Damjan Lampret, lampret@opencores.org
// Description:	
//  OR1001 Integer data path controller. Pipeline stages are fetch, decode,
//  issue, execute, writeback and retire.
// Architecture Notes:	
//  Each pipeline stage has its own process.
// TO DO:
//  * 
//---------------------------------------------------------------------------
// Revision History
// $Log: iuctrl.vhd,v $
// Revision 1.1.1.1  2000/04/18 13:19:39  lampret
// This implementation is now in the middle of modifications to have one CPU core both for OpenRISC 1000 architecture and MIPS I.
//
// Translation to Verilog by Peter Monta <pmonta@terayon.com>, 22 April 2000

`include "config.h"

module iuctrl(fetched, fetch_stall, fetch_biubusy, fetch_branch,
  rfaddr1, rfaddr2, imm, sel1, sel2, lop, cop, aop, save_flag, f_flag,
  rfenw, selwb, rfaddrw, exception, except_type,
  lsu_size, lsu_validload, lsu_validstore, lsu_stall, iu_stall,
  rst, clk);

  parameter datawidth = `DATA_WIDTH;
  parameter insnwidth = `INSN_WIDTH;

// signals related to instruction fetch
  input [insnwidth-1:0] fetched;
  output fetch_stall;
  input fetch_biubusy;
  output fetch_branch;
// signals that control instruction decode stage of integer data path
  output [3:0] rfaddr1;
  output [3:0] rfaddr2;
// signals that control execute stage of integer data path
  output [datawidth-1:0] imm;
  output [1:0] sel1;
  output [1:0] sel2;
  output [1:0] lop;
  output [3:0] cop;
  output aop;
  output save_flag;
  input f_flag;
// signals that control write-back stage of integer data path
  output rfenw;
  output [1:0] selwb;
  output [3:0] rfaddrw;
// exceptions
  output exception;
  output [3:0] except_type;
// branches
// lsu
  output [1:0] lsu_size;
  output lsu_validload;
  output lsu_validstore;
  input lsu_stall;
// iu stall
  output iu_stall;
// reset & clock
  input rst;
  input clk;

  reg [1:0] sel1;
  reg [1:0] sel2;
  reg aop;
  reg [1:0] lop;
  reg [3:0] cop;
  reg [1:0] selwb;
  reg [1:0] lsu_size;
  reg lsu_validload;
  reg lsu_validstore;
  reg fetch_branch;
  reg save_flag;

// instruction decode stage related signals
  reg [`OPC_FLD_SIZE-1:0] id_opcode;
  reg [`REG_FLD_SIZE-1:0] id_reg_d;
  reg [`REG_FLD_SIZE-1:0] id_reg_a;
  reg [`REG_FLD_SIZE-1:0] id_reg_b;
  reg [`IMM_FLD_SIZE-1:0] id_simm;

// instruction issue stage related signals
  reg [`OPC_FLD_SIZE-1:0] is_opcode;
  reg [`REG_FLD_SIZE-1:0] is_reg_d;
  reg [`REG_FLD_SIZE-1:0] is_reg_a;
  reg [`REG_FLD_SIZE-1:0] is_reg_b;
  reg [datawidth-1:0] is_simm;
  reg is_rfenw;

// instruction execute stage related signals
  reg [`OPC_FLD_SIZE-1:0] ex_opcode;
  reg [`REG_FLD_SIZE-1:0] ex_reg_d;
  reg [`REG_FLD_SIZE-1:0] ex_reg_a;
  reg [`REG_FLD_SIZE-1:0] ex_reg_b;
  reg ex_rfenw;

// instruction write-back stage related signals
  reg [`REG_FLD_SIZE-1:0] wb_reg_d;
  reg wb_rfenw;

// stall related signals
  wire if_stall;
  wire id_stall;
  wire is_stall;
  wire ex_stall;
  wire wb_stall;
  wire fetch_stall_out;
  wire datapath_stall;
  wire [`REG_FLD_SIZE+1:0] raw1_compare_a;
  wire [`REG_FLD_SIZE+1:0] raw1_compare_b;
  wire [`REG_FLD_SIZE+1:0] raw2_compare_a;
  wire [`REG_FLD_SIZE+1:0] raw2_compare_b;
  wire raw_hazard1;
  wire raw_hazard2;
  wire raw_hazard;
  reg multicycle_stall;
  reg multicycle_first;

// number of cycles for current insn in EX stage
  reg [4:0] multicycle_cycles;

//
// Instruction fetch stage
//
  always @(posedge clk or posedge rst)
    if (rst) begin
      id_opcode <= #1 0;
      id_reg_d <= #1 0;
      id_reg_a <= #1 0;
      id_reg_b <= #1 0;
      id_simm <= #1 0;
    end else
      if (!if_stall) begin
        id_opcode <= #1 fetched[insnwidth-1:insnwidth-`OPC_FLD_SIZE];
        id_reg_d <= #1 fetched[insnwidth-(`OPC_FLD_SIZE+1):insnwidth-(`OPC_FLD_SIZE+`REG_FLD_SIZE)];
        id_reg_a <= #1 fetched[insnwidth-(`OPC_FLD_SIZE+`REG_FLD_SIZE+1):insnwidth-(`OPC_FLD_SIZE+2*`REG_FLD_SIZE)];
        id_reg_b <= #1 fetched[insnwidth-(`OPC_FLD_SIZE+2*`REG_FLD_SIZE+1):insnwidth-(`OPC_FLD_SIZE+3*`REG_FLD_SIZE)];
        id_simm <= #1 fetched[insnwidth-(`OPC_FLD_SIZE+2*`REG_FLD_SIZE+1):insnwidth-(`OPC_FLD_SIZE+2*`REG_FLD_SIZE+`IMM_FLD_SIZE)];
      end

//
// Instruction decode stage
//
  always @(posedge clk or posedge rst)
    if (rst) begin
      is_opcode <= #1 0;
      is_reg_d <= #1 0;
      is_reg_a <= #1 0;
      is_reg_b <= #1 0;
      is_simm <= #1 0;
      multicycle_cycles <= #1 0;
      multicycle_first <= #1 0;
      sel1 <= #1 0;
      sel2 <= #1 0;
      is_rfenw <= #1 0;
    end else
      if (raw_hazard || fetch_biubusy) begin
// insert bubble
        is_opcode <= #1 `OR32_NOP;
        is_reg_d <= #1 0;
        is_reg_a <= #1 0;
        is_reg_b <= #1 0;
        is_simm <= #1 0;
        multicycle_cycles <= #1 0;
        multicycle_first <= #1 0;
        sel1 <= #1 0;
        sel2 <= #1 0;
        is_rfenw <= #1 0;			
      end else if (!id_stall) begin
        is_opcode <= #1 id_opcode;
        is_reg_d <= #1 id_reg_d;
        is_reg_a <= #1 id_reg_a;
        is_reg_b <= #1 id_reg_b;
// Immediate is not always in the lower 16 bits. If operand B is
// a register then part of immediate is in destination register field
// of instruction
        if (id_opcode[4]) begin
          is_simm[`IMM_FLD_SIZE-1:0] <= #1 {id_reg_d, id_simm[`IMM_FLD_SIZE-`REG_FLD_SIZE-1:0]};
          is_simm[`DATA_WIDTH-1:`IMM_FLD_SIZE] <= #1 {(`DATA_WIDTH-`IMM_FLD_SIZE){id_reg_d[`REG_FLD_SIZE-1]}};
        end else begin
          is_simm[`IMM_FLD_SIZE-1:0] <= #1 id_simm;
          is_simm[`DATA_WIDTH-1:`IMM_FLD_SIZE] <= #1 {(`DATA_WIDTH-`IMM_FLD_SIZE){id_simm[`IMM_FLD_SIZE-1]}};
        end
// forwarding logic for bypassing register file and forwarding
// results from WB and RET stages to source 1 in IS stage
        if (id_opcode[5] && ex_rfenw && (id_reg_a == ex_reg_d))
          sel1 <= #1 `SEL1_FORW_WB;
        else if (id_opcode[5] && wb_rfenw && (id_reg_a == wb_reg_d))
          sel1 <= #1 `SEL1_FORW_RET;
        else
          sel1 <= #1 `SEL1_REGFILE;
// forwarding logic for bypassing register file and forwarding
// results from WB and RET stages to source 2 in IS stage
        if (id_opcode[4] == 0)
          sel2 <= #1 `SEL2_IMM;
	else if (ex_rfenw && (id_reg_b == ex_reg_d))
          sel2 <= #1 `SEL2_FORW_WB;
        else if (wb_rfenw && (id_reg_b == wb_reg_d))
          sel2 <= #1 `SEL2_FORW_RET;
        else
          sel2 <= #1 `SEL2_REGFILE;
// decode is_rfenw
        case (id_opcode)
          `OR32_JAL: begin
                       is_rfenw <= #1 1;
                       multicycle_cycles <= #1 0;
                       multicycle_first <= #1 0;
                     end
          `OR32_BAL: begin
                       is_rfenw <= #1 1;
                       multicycle_cycles <= #1 0;
                       multicycle_first <= #1 0;
                     end
          `OR32_BNF: begin
                       is_rfenw <= #1 0;
                       multicycle_cycles <= #1 0;
                       multicycle_first <= #1 0;
                     end
          `OR32_BF: begin
                      is_rfenw <= #1 0;
                      multicycle_cycles <= #1 0;
                      multicycle_first <= #1 0;
                    end
          `OR32_RFE: begin
                       is_rfenw <= #1 0;
                       multicycle_cycles <= #1 0;
                       multicycle_first <= #1 0;
                     end
          `OR32_SYS: begin
                       is_rfenw <= #1 0;
                       multicycle_cycles <= #1 0;
                       multicycle_first <= #1 0;
                     end
          `OR32_IHU: begin
                       is_rfenw <= #1 1;
                       multicycle_cycles <= #1 0;
                       multicycle_first <= #1 0;
                     end
          `OR32_MFSR: begin
                        is_rfenw <= #1 1;
                        multicycle_cycles <= #1 0;
                        multicycle_first <= #1 0;
                      end
          `OR32_MTSR: begin
                        is_rfenw <= #1 0;
                        multicycle_cycles <= #1 0;
                        multicycle_first <= #1 0;
                      end
          `OR32_LW: begin
                      is_rfenw <= #1 1;
                      multicycle_cycles <= #1 2;
                      multicycle_first <= #1 1;
                    end
          `OR32_LBZ: begin
                       is_rfenw <= #1 1;
                       multicycle_cycles <= #1 2;
                       multicycle_first <= #1 1;
                     end
          `OR32_LBS: begin
                       is_rfenw <= #1 1;
                       multicycle_cycles <= #1 2;
                       multicycle_first <= #1 1;
                     end
          `OR32_LHZ: begin
                       is_rfenw <= #1 1;
                       multicycle_cycles <= #1 2;
                       multicycle_first <= #1 1;
                     end
          `OR32_LHS: begin
                       is_rfenw <= #1 1;
                       multicycle_cycles <= #1 2;
                       multicycle_first <= #1 1;
                     end
          `OR32_ADDI: begin
                        is_rfenw <= #1 1;
                        multicycle_cycles <= #1 0;
                        multicycle_first <= #1 0;
                      end
          `OR32_ADDIC: begin
                         is_rfenw <= #1 1;
                         multicycle_cycles <= #1 0;
                         multicycle_first <= #1 0;
                       end
          `OR32_SUBI: begin
                        is_rfenw <= #1 1;
                        multicycle_cycles <= #1 0;
                        multicycle_first <= #1 0;
                      end
          `OR32_ANDI: begin
                        is_rfenw <= #1 1;
                        multicycle_cycles <= #1 0;
                        multicycle_first <= #1 0;
                      end
          `OR32_ORI: begin
                       is_rfenw <= #1 1;
                       multicycle_cycles <= #1 0;
                       multicycle_first <= #1 0;
                     end
          `OR32_XORI: begin
                        is_rfenw <= #1 1;
                        multicycle_cycles <= #1 0;
                        multicycle_first <= #1 0;
                      end
          `OR32_SH_ROT: begin
                          is_rfenw <= #1 1;
                          multicycle_cycles <= #1 1;
                          multicycle_first <= #1 1;
                        end
          `OR32_SFXXI: begin
                         is_rfenw <= #1 0;
                         multicycle_cycles <= #1 0;
                         multicycle_first <= #1 0;
                       end
          `OR32_SW: begin
                      is_rfenw <= #1 0;
                      multicycle_cycles <= #1 1;
                      multicycle_first <= #1 1;
                    end
          `OR32_SB: begin
                      is_rfenw <= #1 0;
                      multicycle_cycles <= #1 1;
                      multicycle_first <= #1 1;
                    end
          `OR32_SH: begin
                      is_rfenw <= #1 0;
                      multicycle_cycles <= #1 1;
                      multicycle_first <= #1 1;
                    end
          `OR32_ALU: begin
                       is_rfenw <= #1 1;
                       multicycle_cycles <= #1 0;
                       multicycle_first <= #1 0;
                     end
          `OR32_SFXX: begin
                        is_rfenw <= #1 0;
                        multicycle_cycles <= #1 0;
                        multicycle_first <= #1 0;
                      end
          default: begin
                     is_rfenw <= #1 0;
                     multicycle_cycles <= #1 0;
                     multicycle_first <= #1 0;
                   end
        endcase
      end

//
// Instruction issue stage
//
  always @(posedge clk or posedge rst)
    if (rst) begin
      ex_opcode <= #1 0;
      ex_reg_d <= #1 0;
      ex_reg_a <= #1 0;
      ex_reg_b <= #1 0;
      ex_rfenw <= #1 0;
      aop <= `AOP_X;
      lop <= `LOP_X;
      lsu_validload <= #1 0;
      lsu_validstore <= #1 0;
      lsu_size <= #1 `LSU_SIZE_X;
      fetch_branch <= #1 0;
      save_flag <= #1 0;
    end else
      if (!is_stall) begin
        ex_opcode <= #1 is_opcode;
        ex_reg_d <= #1 is_reg_d;
        ex_reg_a <= #1 is_reg_a;
        ex_reg_b <= #1 is_reg_b;
        ex_rfenw <= #1 is_rfenw;
        cop <= #1 is_reg_d[3:0];
        case (is_opcode)
          `OR32_JAL: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 1;
                       save_flag <= #1 0;
                     end
          `OR32_BAL: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_WORD;
                       fetch_branch <= #1 1;
                       save_flag <= #1 0;
                     end
          `OR32_BNF: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 !f_flag;
                       save_flag <= #1 0;
                     end
          `OR32_BF: begin
                      aop <= #1 `AOP_X;
                      lop <= #1 `LOP_X;
                      lsu_validload <= #1 0;
                      lsu_validstore <= #1 0;
                      lsu_size <= #1 `LSU_SIZE_X;
                      fetch_branch <= #1 f_flag;
                      save_flag <= #1 0;
                    end
          `OR32_RFE: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 1;
                       save_flag <= #1 0;
                     end
          `OR32_SYS: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_IHU: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_MFSR: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_MTSR: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_LW: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 1;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_WORD;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_LBZ: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 1;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_BYTE;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_LBS: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 1;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_BYTE;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_LHZ: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 1;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_HALF;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_LHS: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 1;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_HALF;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_ADDI: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_ADDIC: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_SUBI: begin
                       aop <= #1 `AOP_SUB;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_ANDI: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_AND;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_ORI: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_OR;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_XORI: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_XOR;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_SH_ROT: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_SFXXI: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_AND;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 1;
                     end
          `OR32_SW: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 1;
                       lsu_size <= #1 `LSU_SIZE_WORD;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_SB: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 1;
                       lsu_size <= #1 `LSU_SIZE_BYTE;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_SH: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 1;
                       lsu_size <= #1 `LSU_SIZE_HALF;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_ALU: begin
                       aop <= #1 `AOP_ADD;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
          `OR32_SFXX: begin
                       aop <= #1 `AOP_SUB;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 1;
                     end
          default: begin
                       aop <= #1 `AOP_X;
                       lop <= #1 `LOP_X;
                       lsu_validload <= #1 0;
                       lsu_validstore <= #1 0;
                       lsu_size <= #1 `LSU_SIZE_X;
                       fetch_branch <= #1 0;
                       save_flag <= #1 0;
                     end
        endcase
      end

//
// Instruction execute stage
//
  always @(posedge clk or posedge rst)
    if (rst) begin
      wb_reg_d <= #1 0;
      wb_rfenw <= #1 0;
      selwb <= #1 `SELWB_X;
    end else
      if (!ex_stall) begin
        wb_reg_d <= #1 ex_reg_d;
        wb_rfenw <= #1 ex_rfenw;
        case (ex_opcode)
          `OR32_JAL: selwb <= #1 `SELWB_LSU;
          `OR32_BAL: selwb <= #1 `SELWB_LSU;
          `OR32_BNF: selwb <= #1 `SELWB_X;
          `OR32_BF: selwb <= #1 `SELWB_X;
          `OR32_RFE: selwb <= #1 `SELWB_X;
          `OR32_SYS: selwb <= #1 `SELWB_X;
          `OR32_IHU: selwb <= #1 `SELWB_ARITH;
          `OR32_MFSR: selwb <= #1 `SELWB_LSU;
          `OR32_MTSR: selwb <= #1 `SELWB_X;
          `OR32_LW: selwb <= #1 `SELWB_LSU;
          `OR32_LBZ: selwb <= #1 `SELWB_LSU;
          `OR32_LBS: selwb <= #1 `SELWB_LSU;
          `OR32_LHZ: selwb <= #1 `SELWB_LSU;
          `OR32_LHS: selwb <= #1 `SELWB_LSU;
          `OR32_ADDI: selwb <= #1 `SELWB_ARITH;
          `OR32_ADDIC: selwb <= #1 `SELWB_ARITH;
          `OR32_SUBI: selwb <= #1 `SELWB_ARITH;
          `OR32_ANDI: selwb <= #1 `SELWB_LOGIC;
          `OR32_ORI: selwb <= #1 `SELWB_LOGIC;
          `OR32_XORI: selwb <= #1 `SELWB_LOGIC;
          `OR32_SH_ROT: selwb <= #1 `SELWB_SH_ROT;
          `OR32_SFXXI: selwb <= #1 `SELWB_X;
          `OR32_SW: selwb <= #1 `SELWB_X;
          `OR32_SB: selwb <= #1 `SELWB_X;
          `OR32_SH: selwb <= #1 `SELWB_X;
          `OR32_ALU: selwb <= #1 `SELWB_ARITH;
          `OR32_SFXX: selwb <= #1 `SELWB_X;
          default: selwb <= #1 `SELWB_X;
        endcase
      end

//
// Connect pipeline signals to register file signals
//
// OR1001 has only lower 16 GPRs
  assign rfaddr1 = id_reg_a[3:0];
  assign rfaddr2 = id_reg_b[3:0];
  assign rfaddrw = wb_reg_d[3:0];
  assign rfenw = wb_rfenw;

//
// Detect RAW hazards
//
// check for RAW and if it exists insert a bubble in IS stage
  assign raw_hazard = (id_opcode[5] && is_rfenw && id_reg_a==is_reg_d) ||
                      (id_opcode[4] && is_rfenw && id_reg_b==is_reg_d);

//pragma translate_off
//raw1_compare_a <= id_opcode(5) & is_rfenw & id_reg_a;
//raw1_compare_b <= vcc & vcc & is_reg_d;
//
//raw2_compare_a <= id_opcode(4) & is_rfenw & id_reg_b;
//raw2_compare_b <= vcc & vcc & is_reg_d;
//
//raw1_detect : comp
//generic map ( width => (REG_FLD_SIZE + 2))
//port map (
//	a => raw1_compare_a,
//	b => raw1_compare_b,
//	eq_in => vcc,
//	eq => raw_hazard1
//);
//
//raw2_detect : comp
//generic map ( width => (REG_FLD_SIZE + 2))
//port map (
//	a => raw2_compare_a,
//	b => raw2_compare_b,
//	eq_in => vcc,
//	eq => raw_hazard2
//);
//
//raw_hazard <= raw_hazard1 or raw_hazard2;
//pragma translate_on
//raw_hazard <= 0;

//
// Multicycle operation pipeline stall
//
  reg [4:0] multicycle_cnt;

  always @(posedge clk or posedge rst)
    if (rst) begin
      multicycle_cnt <= #1 0;
      multicycle_stall <= #1 1;
    end else begin
      if (multicycle_cnt==0)
        multicycle_cnt <= #1 multicycle_cycles;
      if (multicycle_cnt>0) begin
        multicycle_cnt <= #1 multicycle_cnt - 1;
        multicycle_stall <= #1 1;
      end else
        multicycle_stall <= #1 0;
    end

//
// Generating different stall signals
//

// IF stage stall (writing to id_ registers)
  assign if_stall = multicycle_stall || fetch_biubusy || raw_hazard;

// ID stage stall (writing to is_ registers)
  assign id_stall = multicycle_stall || fetch_biubusy || raw_hazard;

// IS stage stall (writing to ex_ registers)
  assign is_stall = multicycle_stall || fetch_biubusy;

// EX stage stall (writing to wb_ registers)
  assign ex_stall = multicycle_stall || fetch_biubusy;

// WB stage stall (writing to ret_ registers)
  assign wb_stall = multicycle_stall || fetch_biubusy;
	
// stall fetcher on all type of hazards
  always @(posedge clk)
    $display($time,,"fetch_stall %x %x %x %x",multicycle_stall,fetch_biubusy,raw_hazard,multicycle_first);
  assign fetch_stall_out = multicycle_stall || fetch_biubusy ||
    raw_hazard || multicycle_first;

// stall datapath IS,EX,WB stages if multicycle stall
  assign datapath_stall = multicycle_stall;

  assign exception = 0;
  assign except_type = 4'b0000;
  assign imm = is_simm;
  assign fetch_stall = fetch_stall_out;
  assign iu_stall = datapath_stall;
   
endmodule
